#define BLOCK_SIZE 256

#include <algorithm>
#include <iostream>
#include <random>
#include <string>
#include <vector>

#include <cuda_profiler_api.h>

__global__ void count(int const *labels, int *counts, int width)
{
    int x = blockIdx.x * blockDim.x + threadIdx.x;

    if (x >= width)
        return;

    int label = labels[x];
    atomicAdd(counts + label, 1);
}

std::vector<int> generateData(int numValues)
{
    std::mt19937 rng;
    std::uniform_int_distribution<int> dist(0, numValues / 10);

    auto generateNumber = [&]() {
        return dist(rng);
    };

    std::vector<int> data(numValues);
    std::generate(data.begin(), data.end(), generateNumber);

    return data;
}

void testIteration()
{
    int const width = 1280 * 1024;
    std::vector<int> data = generateData(width);

    int *labels;
    cudaMalloc(&labels, width * sizeof(int));
    cudaMemcpy(labels, data.data(), width * sizeof(int), cudaMemcpyHostToDevice);

    int *counts;
    cudaMalloc(&counts, width * sizeof(int));
    cudaMemset(&counts, 0, width * sizeof(int));

    dim3 blocks((width + BLOCK_SIZE - 1) / BLOCK_SIZE);
    dim3 threads(BLOCK_SIZE);
    count<<<blocks, threads>>>(labels, counts, width);
    cudaDeviceSynchronize();

    /*cudaMemcpy(data.data(), counts, width * sizeof(int), cudaMemcpyDeviceToHost);
    for (int i = 0; i < width / 10; i++) {
        std::cout << data[i] << "\n";
    }
    std::cout << std::endl;*/

    cudaFree(labels);
    cudaFree(counts);
}

int main(int argc, char **argv)
{
    int iterations = 5;
    if (argc > 1)
    {
        iterations = std::stoi(argv[1]);
    }

    cudaProfilerStart();

    for (int i = 0; i < iterations; i++)
    {
        testIteration();
    }

    cudaProfilerStop();

    return 0;
}
